
// aiguille et zone

#define AIGUILLE_DEVIEE false
#define AIGUILLE_DIRECTE true

#define ZONE_LIBRE false
#define ZONE_OCCUPEE true

/////////////////////////////////////////////////////////////////////////////

class Aiguille { 
protected:
  boolean etat; // directe (true) ou deviee (false)
public: 
  int no;       // numéro de l'aiguille pour commander l'electronique (moteur)
  char* nom;    // le nom de l'aiguille   OPTION nom
  Zone* zone;   // la Zone ou est l'aiguille   OPTION zone
  
  Aiguille(int n) { no=n; } // constructeur mini
  Aiguille(char* nm,int n) { nom=nm; no=n; } // constructeur maxi
  Aiguille(char* nm,int n,Zone* z) { nom=nm; no=n; zone=z; } // constructeur OPTION

  boolean directe() { return etat; }  // méthode de test
  boolean deviee()  { return !etat; } // méthode de test
  
  boolean directer(); // mettre l'aiguille en position directe
  boolean devier();   // mettre l'aiguille en position deviee

  boolean manoeuvrable(); // OPTION zone
  boolean manoeuvrer(); // commande effective de l'aiguille
  
  void init1(boolean e) { etat=e; } // initialisation etat
  void init2(boolean e) { manoeuvrer(); etat=e; } // mise en position initiale
  
  //void erreur(); // erreur sur les aiguilles (tranche 100)
};

/////////////////////////////////////////////////////////////////////////////

enum { 
// zones (sens de circulation possibles)   OPTION
  SENS_IMPAIR_SEUL=1,         // sens impair uniquement
  SENS_PAIR_SEUL=2,           // sens pair uniquement
  SENS_IMPAIR_OU_PAIR=4,      // impair ou pair sans changement possible
  SENS_IMPAIR_ET_PAIR=8,      // impair ou pair avec changement possible
// autres caracteristiques possibles 
// LIGNE,GARE,LONGUE,COURTE,RETROSIGNALISATION,PLEINE_VOIE,TIROIR,AIGUILLES,QUAI,FOND,DEVANT, ...
};

class Zone { 
protected:
   boolean etat; // libre (false) ou occupé (true)
public:
   int no; // le numero de la zone pour commander l'electronique ( TCO, relais, ... )
   char* nom; // le nom de la zone   OPTION
   int type;  // caracteristiques de la zone   OPTION
   Signal* signalPair; Signal* signalImpair; // les signaux eventuels de la zone   OPTION
protected:   
   Zone(int n) { no=n; } // constructeur mini
   Zone(char* nm,int t,int n) { nom=nm, type=t; no=n; } // constructeur maxi
   Zone(char* nm,int t,int n,Signal* sp,Signal* si):Zone(nm,t,n) { signalPair=sp; signalImpair=si; } // constructeur   OPTION
public: 
   boolean occupee() { return etat; }  // méthode de test
   boolean libre()   { return !etat; } // méthode de test

   virtual Zone* suivantePaire()   { return NULL; } // la zone suivante paire (éventuellement vide)   OPTION
   virtual Zone* suivanteImpaire() { return NULL; } // la zone suivante impaire (éventuellement vide)   OPTION

   virtual void actions() {} // les actions spécifiques à faire en cas d'occupation
   virtual void desactions() {} // les actions spécifiques à faire en cas de libération

   void occuper(); // appelée par la rétrosignalisation lors de l'occupation d'une zone
   void liberer(); // appelée par la rétrosignalisation lors de la liberation d'une zone

   void init(boolean b) { etat=b; } // initialisation de l'etat


// méthodes utilitaires

   Zone* selonAiguille(Aiguille* a,Zone* z1,Zone* z2);
   
   Zone* provenance(); // de quelle zone le train vient lors d'une occupation de zone   OPTION
   
   void erreur(int n); // erreur sur les zones (tranche 200)
};

/////////////////////////////////////////////////////////////////////////////

inline boolean Aiguille::manoeuvrable() { return zone->libre(); } // OPTION zone

boolean Aiguille::directer() { if (directe()) return true; if (!manoeuvrer()) return false; etat=AIGUILLE_DIRECTE; return true; }

boolean Aiguille::devier()   { if (deviee())  return true; if (!manoeuvrer())  return false; etat=AIGUILLE_DEVIEE;  return true; }

boolean Aiguille::manoeuvrer() { // appele quand une aiguille est manoeuvree
  // COMPLETER mettre ici le code qui manoeuvre effectivement l'aiguille ( utiliser le booleen etat et l'entier no )
  return true; // return false si cela c'est mal passe
}

//void  Aiguille::erreur() {}

//  //  // //  //  //  //  //  //  //  //  //  //  //  //  //  //  //  //  //

void Zone::occuper() { // appelée par la rétrosignalisation
  if (etat==ZONE_OCCUPEE) return; etat=ZONE_OCCUPEE;
  actions(); // fait les actions spécifiques à une zone
  // COMPLETER fait tout ce qu'il y a à faire en cas d'occupation (actions communes à toutes les zones)
  // COMPLETER fait les actions de l'utilisateur (TCO)
}
void Zone::liberer() { // appelée par la rétrosignalisation
  if (etat==ZONE_LIBRE) return; etat=ZONE_LIBRE;
  desactions(); // fait les actions spécifiques à une zone
  // COMPLETER fait tout ce qu'il y a à faire en cas de libération (actions communes à toutes les zones)
  // COMPLETER fait les actions de l'utilisateur (TCO)
}

Zone* Zone::provenance() { Zone* p; Zone* i; // OPTION
   // if (type==SENS_IMPAIR_SEUL) return suivantePaire();   // optimisation (si type)
   // if (type==SENS_PAIR_SEUL)   return suivanteImpaire(); // optimisation (si type)
   p=suivantePaire();   if (p!=NULL && p->libre()) p=NULL; // candidate si non nulle et occupee
   i=suivanteImpaire(); if (i!=NULL && i->libre()) i=NULL; // candidate si non nulle et occupee
   if (p!=NULL) if (i==NULL) return p; else erreur(201); // deux candidates (ambiguite)
   else         if (i!=NULL) return i; else erreur(202); // aucune candidate (erreur)
}

inline Zone* Zone::selonAiguille(Aiguille* a,Zone* z1,Zone* z2) { return a->directe()?z1:z2; } 
 
void Zone::erreur(int n) { /* COMPLETER */ }

/////////////////////////////////////////////////////////////////////////////

